%  Copyright 2015 Kwang In Kim, Max-Planck-Institut for Informatics
%  Please refer to license.txt for sanctioned usage.
%
% [L] = ConstructLap(Nodes, NNIdx, SigmaSq)
% INPUT:
% 'Nodes': (N times d) data matrix; 'N': number of data points, 'd': 
%           dimensionality of data.
% 'NNIdx': (N times k) matrix containing indices of k-nearest neighbors for
%           each data point (This matrix can be obtained using GetKNN).
% 'SigmaSq': the 'usual' SigmaSq parameter for the similarity matrix used in building graph Laplacian.
%
% OUTPUT:
% 'L': (N times N) matrix - (unnormalized) grpah Laplacian matrix L.

function [L] = ConstructLap(Nodes, NNIdx, SigmaSq)

[NumNodes,AmbDim] = size(Nodes);

%%% sparse version
    BlSpMatSize = 300000;
    MaxNumNN = size(NNIdx,2);
    NumNonzero = NumNodes*MaxNumNN;
    BlSpMatNum = ceil(NumNonzero/BlSpMatSize);
    BlSpL = cell(BlSpMatNum,1);
    BlSpCtVal = cell(BlSpMatNum,1);
%%% sparse version

SpMatIdx = 1;
SpEleCounter = 1;
SpRowIdx = zeros(BlSpMatSize,1);
SpColIdx = zeros(BlSpMatSize,1);
LVal = zeros(BlSpMatSize,1);
CtVal = zeros(BlSpMatSize,1);
for i=1:NumNodes
    G = Nodes(NNIdx(i,:),:);
    [NumNeighbors,AmbDim] = size(G);
    if nargin < 2
        SigmaSq = -1;
    end
    
    [LocalL] = ComputeGaussianKernel(G, SigmaSq, Nodes(i,:));
    
    GRIdx = i;
    GCIdx = NNIdx(i,1:NumNeighbors);
    SpRowIdx(SpEleCounter:SpEleCounter+NumNeighbors-1) = GRIdx;
    SpColIdx(SpEleCounter:SpEleCounter+NumNeighbors-1) = GCIdx;
    LVal(SpEleCounter:SpEleCounter+NumNeighbors-1) = LocalL(1:NumNeighbors);
    CtVal(SpEleCounter:SpEleCounter+NumNeighbors-1) = CtVal(SpEleCounter:SpEleCounter+NumNeighbors-1)+1;
    SpEleCounter = SpEleCounter+NumNeighbors;

    if SpEleCounter > BlSpMatSize-NumNeighbors-1
        NonZeroIdx = length(find(SpRowIdx>0));
        SpRowIdx = SpRowIdx(1:NonZeroIdx);
        SpColIdx = SpColIdx(1:NonZeroIdx);
        LVal = LVal(1:NonZeroIdx);
        CtVal = CtVal(1:NonZeroIdx);
        BlSpL{SpMatIdx} = sparse(SpRowIdx,SpColIdx,LVal,NumNodes,NumNodes);
        BlSpCtVal{SpMatIdx} = sparse(SpRowIdx,SpColIdx,CtVal,NumNodes,NumNodes);
        
        SpMatIdx = SpMatIdx+1;
        SpEleCounter = 1;
        SpRowIdx = zeros(BlSpMatSize,1);
        SpColIdx = zeros(BlSpMatSize,1);
        LVal = zeros(BlSpMatSize,1);
        CtVal = zeros(BlSpMatSize,1);
    end
end

if SpEleCounter > 1
    NonZeroIdx = length(find(SpRowIdx>0));
    SpRowIdx = SpRowIdx(1:NonZeroIdx);
    SpColIdx = SpColIdx(1:NonZeroIdx);
    LVal = LVal(1:NonZeroIdx);
    CtVal = CtVal(1:NonZeroIdx);
    BlSpL{SpMatIdx} = sparse(SpRowIdx,SpColIdx,LVal,NumNodes,NumNodes);
    BlSpCtVal{SpMatIdx} = sparse(SpRowIdx,SpColIdx,CtVal,NumNodes,NumNodes);
        
    SpMatIdx = SpMatIdx+1;
end

L = sparse([],[],[],NumNodes,NumNodes,0);
CtValMat = sparse([],[],[],NumNodes,NumNodes,0);
for i=1:SpMatIdx-1
    L = L+BlSpL{i};
    CtValMat = CtValMat+BlSpCtVal{i};
end
%%% symmetrize nn structure
    L = (L'+L)/2;
    CtIdx = find(CtValMat>1);
    L(CtIdx) = L(CtIdx)./CtValMat(CtIdx);
%%% symmetrize nn structure

SpRowIdx = [1:NumNodes];
SpColIdx = [1:NumNodes];
DVal = sum(L,2);
D = sparse(SpRowIdx,SpColIdx,DVal,NumNodes,NumNodes);

L = D-L;
L = (L'+L)/2;
